from .registration.decorator import Tktl  # noqa: F401
from .registration.endpoints import (  # noqa: F401
    ArrowEndpoint,
    BasicEndpoint,
    Endpoint,
    ProfiledEndpoint,
    TypedEndpoint,
)
