class CLIStrings:
    INVALID_INPUT = "Invalid Input: "
    AUTH_ERROR_MISSING_KEY = "No keys found locally. Please login first."
    REQUEST_FAILED = "Request Failed: "


class HeaderStrings:
    APPLICATION_JSON = "application/json"
    APPLICATION_URLENCODED = "application/x-www-form-urlencoded"
